'''
Module which implements the synapse module API/convention.
'''
coremods = (
    'synapse.models.dns.DnsModule',
    'synapse.models.orgs.OuModule',
    'synapse.models.syn.SynModule',
    'synapse.models.auth.AuthModule',
    'synapse.models.base.BaseModule',
    'synapse.models.risk.RiskModule',
    'synapse.models.person.PsModule',
    'synapse.models.files.FileModule',
    'synapse.models.infotech.ItModule',
    'synapse.models.geospace.GeoModule',
    'synapse.models.media.MediaModule',
    'synapse.models.geopol.PolModule',
    'synapse.models.telco.TelcoModule',
    'synapse.models.inet.InetModule',
    'synapse.models.material.MatModule',
    'synapse.models.language.LangModule',
    'synapse.models.crypto.CryptoModule',
    'synapse.models.gov.cn.GovCnModule',
    'synapse.models.gov.us.GovUsModule',
    'synapse.models.gov.intl.GovIntlModule',
    'synapse.models.economic.EconModule',
    'synapse.models.transport.TransportModule',
    'synapse.models.proj.ProjectModule',
    'synapse.models.biz.BizModule',
)
