"""High-level XRPL exports."""
#
# Copyright (c) 2021, XRP Ledger Foundation
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
# FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
# OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# -----------------------------------------------------------------------------------------------------------------------0
#
# The vendored library is a subset of the xrpl-py v1.1.1 library for the purposes of Cryptocurrency address validation.
# It has been modified for vendored imports.
# '''

from synapse.vendor.xrpl import core
from synapse.vendor.xrpl.constants import CryptoAlgorithm, XRPLException

__all__ = [
    "CryptoAlgorithm",
    "XRPLException",
    "core",
]
