"""General XRPL Address Codec Exceptions."""
# It has been modified for vendored imports.
from synapse.vendor.xrpl.constants import XRPLException


class XRPLAddressCodecException(XRPLException):
    """General XRPL Address Codec Exception."""

    pass
