import synapse.axon as s_axon
import synapse.cortex as s_cortex
import synapse.cryotank as s_cryotank

import synapse.lib.aha as s_aha
import synapse.lib.jsonstor as s_jsonstor

aha = s_aha.AhaCell
axon = s_axon.Axon
cortex = s_cortex.Cortex
cryotank = s_cryotank.CryoCell
jsonstor = s_jsonstor.JsonStorCell
